""" Functions this module are used to read model (GEOS-Chem data) and  
regrid them to the given h/v resolution 

    Functions:
    ===================================================
    1. get_model_pressure:calculate pressure at model level from surface pressure
    2. get_model_value: read values from bpch2 file 
    3. regrid_data:regrid data 
    4. get_model_zonal_mean: calculate model zonal mean field 
    5. get_model_colum: calculate model column 
    

"""

# from pylab import *

# try: 
#     from matplotlib.toolkits.basemap import Basemap, shiftgrid
# except ImportError:
#    from mpl_toolkits.basemap import Basemap, shiftgrid

from numpy import * 
import ESA.util.time_module as tm
import ESA.util.gp_axis_m as gax
import ESA.util.flib as flib
import ESA.util.otool_ncfile_io as ncfio
import ESA.atmosphere.compute_gc_grid as cmg
import ESA.util.bpch2_rw_smp as brw
import ESA.util.pres_m as pm

def_pres_file_varnames=['longitude', 'latitude', 'sp']


def get_model_pressure(lx, mod_type=5, ps=None, \
                           psfile=None, cater_ps='PS',\
                           tracer_ps=None, tau_ps=None, \
                           nc_varnames=def_pres_file_varnames):
    
    """
    calculate pressure at model level from surface pressure
    
    Inputs:
    ------------------------------------------------------
    1. lx:<list, t:int>: level index
    2. mod_type:<int>: model type
    3. ps:<array (nlon, nlat)>: surface pressure, if ps is not given, 
    the code will try to read it from netcdf or bpch2 file 
    4. psfile:<str>: name of bpch2 or netcdf file for pressure
    5. cater_ps:<str>: category of surface pressure
    6. tracer_ps:<int>: ID for surface pressure 
    7. tau_ps:<float>:  time for surface pressure
    8. nc_varnames:<list>: name of longitude, latitude, and surface pressure in netcdf file 

    Returns
    =======================================================
    1. pres:<array, (nlon, nlat, nz)>: model pressure
    2. ps:<array, (nlon, nlat)>: surface pressure
    
    """
    # S1: load surface pressure 
    if (psfile<>None):
        
        if ('.nc' in psfile): # the netcdf file
            # T1: read from netCDF file 
            grd_lon, grd_lat, ps=ncfio.ncf_read(psfile, varnames)
            ps=squeeze(ps)
            
        else:
            
            # T1: read from GEOS-Chem output
            category_out,tracer_out,unit_out,\
                tau0_out,out_array= \
                get_model_value(psfile, category=cater_ps, tracer=tracer_ps, tau0=tau_ps)
            print 'get_model_pressure: unit for ps', unit_out.strip()
            
            ps=squeeze(out_array)
            print 'get_model_pressure: shape of ps, min(ps), max(ps)', shape(ps), min(ps.flat), max(ps.flat)
            
    
    
    levels=arange(lx)
    levels=levels+1
    
    if (mod_type==4):  # GEOS-4
        if (lx==30):
            use_reduced=1
        else:                 
            use_reduced=0

        ap, bp=pm.get_geos4_ap_bp(reduced_grid=use_reduced)

    else:                    # GEOS-5
        
        if (lx<50):
            use_reduced=1
        else:
            use_reduced=0

        ap, bp=pm.get_geos5_ap_bp(reduced_grid=use_reduced)
    # 
    pres = pm.get_mod_pres(ps, ap, bp)
    print shape(pres), shape(ps)
    
    
    return pres, ps

def get_model_value(filename, category=None, tracer=None, tau0=None):
    """ this is wrapper for function read_bpch2_firstmatch in bpch2_rw_py for reading model values
    
    Inputs:
    =========================================================
    1. filename:<str>: file name 
    2. category:<str>: category 
    3. tracer:<int>: tracer ID
    4. tau0:<float>: time 
    
    Returns
    =============================================
    1. category_out:<str>: category 
    2. tracer_out:<int>: tracer ID 
    3. unit_out:<str>: unit 
    4. tau0_out:<float>: time 
    5. out_array:<array>: data read from bpch2 file 
    
    """

    import ESA.util.bpch2_rw_py as bp
    
    
    if (category==None):
        category_in='NONE'
    else:
        category_in=category.strip()
        
    if (tracer==None):
        tracer_in=-999
    else:
        tracer_in=int(tracer)
    
    if (tau0==None):
        tau0_in=-999.0
    else:
        tau0_in=tau0
    
    # print 'I am here'
    # print category_in, tracer_in, tau0_in

    print '===== get_model_value===='
    short_flnm=filename.strip()
    print ' read data from:'+short_flnm
    category_out,tracer_out,unit_out,\
        tau0_out,ix,jx,lx,out_array,stat = \
        bp.read_bpch2_firstmatch(short_flnm,category_in,tracer_in,tau0_in)
    
    
    if (stat <> 0):
        print 'error in reading '+flnm, state
        return none

    print 'read_model_value, ix jx lx in read out', ix, jx, lx
    
    out_array=out_array[0:ix, 0:jx, 0:lx]
    print '=====END get_model_value END ===='
    return category_out,tracer_out,unit_out,\
        tau0_out,out_array


def regrid_data(data, lon, lat, \
                    new_lon, new_lat, \
                    pressure=None, new_pres=None,\
                    do_massconservation=True):
    
    """ regrid data 
    
    Inputs:
    ------------------------------------------
    1. data:<array, (nlon, nlat, nz)>: 3D field
    2. lon:<array, (nlon)>: longitude
    3. lat:<array, (nlat)>: latitude
    4. new_lon:<array, (new_nlon)>: new longitude grid 
    5. new_lat:<array, (new_nlat)>: new latitude grid
    4. pressure:<array, (nlon, nlat, nz)>: pressure at each model level
    5. new_pres: <array, (new_nlon, new_nlat, new_nz)>: pressure at new model grid
    
    Returns:
    -----------------------------------------------------------------------
    1. new_data: <array, (new_nlon, new_nlat, new_nz)>: data at new model level
    
    """
    
    
    # S1: horizontal interpolation 
    
    ax_lat=gax.gp_axis_cl('lat', lat)
    ax_lon=gax.gp_axis_cl('lat', lon)

    lonp1, lonp2, lonw=ax_lon.getwgt(new_lon)
    latp1, latp2, latw=ax_lat.getwgt(new_lat)
    
    new_data1=lonw[:,newaxis, newaxis]*data[lonp1, :, :]+\
        (1.0-lonw[:,newaxis, newaxis])*data[lonp2,:,:]
    new_data=latw[newaxis, :, newaxis]*new_data1[:,latp1, :]+\
        (1.0-latw[newaxis, :, newaxis])*new_data1[:,latp2, :]
    
    # print 'new_z in regrid,old_mod, new_ond', size(new_z), old_mod, new_mod
    # do vertical interpolation if required 
    
    old_nx, old_ny, old_nz=shape(data)
    print 'regrid_data---original shape:', shape(data)
    print 'regrid_data---new shape (after horizontal regrid):', shape(new_data)
    
    
    # S2: vertical interpolation if necessary
    
    if (old_nz==1):
        return new_data
    
    
    if ((new_pres<>None) and (pressure<>None)):
        # T1: interpolate pressure to new horizontal grid 
        print 'regrid_data: try to regrid  vertical level'
        
        old_pres1=lonw[:,newaxis, newaxis]*pressure[lonp1, :, :]+\
            (1.0-lonw[:,newaxis, newaxis])*pressure[lonp2,:,:]
        old_pres=latw[newaxis, :, newaxis]*old_pres1[:,latp1, :]+\
            (1.0-latw[newaxis, :, newaxis])*old_pres1[:,latp2, :]
        print 'regrid_data--old_pres.shape after regrid: ', shape(old_pres)
        
        # T2: get vertical interpolation weight 
        old_pres=squeeze(old_pres)
        old_pres_log=-log10(old_pres)
        new_pres_log=-log10(new_pres)
        pl, pr,wgt=flib.get_vertical_wgt_2d(old_pres_log,new_pres_log)
        
        # T3: do vertical interpolation 
        
        prof=new_data
        out_prof = flib.prof_vertical_intpl_2d(pl,pr,wgt,prof)
        new_data=array(out_prof)
        if (do_massconservation):
            new_col=get_model_colum(new_data, new_pres)
            old_col=get_model_colum(prof, old_pres)
            factor=old_col/new_col
            new_data=new_data[:,:,:]*factor[:,:,newaxis]
    
    print 'regrid_data---new shape after regrid:', shape(new_data)
    
    return new_data
    
            
                    
        
def get_model_zonal_mean(data, pressure, pres_lvl):
    """
    calculate model zonal mean 
    Inputs:
    ==========================================
    1.data:<array, (nlon, nlat, nz)>: mixing ratio 
    2. pressure:<array, (nlon, nlat, nz)>: model pressure
    3. pres_lvl:<array, (nzx)>: pressure levels for zonal mean 
    
    
    """

    print '----in get_model_zonal_mean-----'
    print 'shape(data), shape(pressure), shape(pres_lvl):',  \
        shape(data), shape(pressure), shape(pres_lvl)
    
    nz_data=size(data[0,0,:])
    nz_pres=size(pressure[0,0,:])
    nz=min(nz_data, nz_pres)
    data=data[:,:,0:nz]
    pressure=pressure[:,:,0:nz]
    
    vals=flib.get_zonal_mean(data, pressure, pres_lvl)

    print 'shape vals', shape(vals)
    print 'vals[30,0], [30, 10]', vals[30, 0], vals[30, 10]
    print '----leave get_model_zonal_mean-----'
    
    return vals

if (__name__=='__main__'):
    psfile="/geos/u23/enkf_std/enkf_output_22/ts.EN0117-EN0139.20030319.bpch"
    lx=30
    
    pressure=get_model_pressure(lx, mod_type=4, ps=None, psfile=psfile, \
                                cater_ps='TIME-SER',  \
                                tracer_ps=15, tau_ps=None)
    print shape(pressure)
    
def get_model_colum(data, pressure):
    """
    calculate model column values  
    Inputs:
    ==========================================
    1.data:<array, (nlon, nlat, nz)>: mixing ratio 
    2. pressure:<array, (nlon, nlat, nz)>: model pressure

    returns:
    =====================================
    1. colval:<array, (nlon, nlat)>: total column 
    
    """
    # print '----in get_model_column-----'
    # print 'shape(data), shape(pressure)',  shape(data), shape(pressure)
    

    nz_data=size(data[0,0,:])
    nz_pres=size(pressure[0,0,:])
    nz=min(nz_data, nz_pres)
    data=data[:,:,0:nz]
    pressure=pressure[:,:,0:nz]
    
    colwgt=flib.get_col_wgt_2d(pressure)
    colval=flib.col_int_2d(data, colwgt)
    # print 'shape vals', shape(colval)
    # print '----leave get_model_column-----'
    
    return colval


