"""find EOFs for a array 

"""

import numpy as npy
import numpy.linalg as alg
import pylab as plt
import numpy.ma as ma
import ESA.util.gen_plots as gpl

class eof_cl:

    """
    class for solve eof of array
    
    Members:
    ---------------------------------------------------------------------
    
    1. data:<array, (nt, nx, (ny),...)>: original temporal spatial data
    2. wgt:<array, (nt, , nx, (ny),...)>: weigthing factor

    3. usd_data:<array, (nt, usd_np>:  copy of the original data after reshaping and removing bad points
    4. usd_wgt:<array, (nt, usd_np )>: copy of weigthing factor

    5. nw:<int>: number of eigenvalues from SVD decomposition 
    
    6. u:<array, (nt, nw)>: U matrix after SVD decomposition of usd_data
    6. w:<array, (nw)>: eigenvalues from  SVD decomposition of usd_data
    7. vt:<array, (nw, usd_np)>: transpose of eigenstate of error covariances
    8. do_reshape:<T/F>: True, if reshaping is necessary 
    9. filling_val:<float>: filling values
    10. usd_idx:<list>: index for points have been used to derive EOFs
    11. dims:<list>: shape of data
    12. ndims:<int>: number of dimensions
    13. pcs:<array, (nt, nw)>: Principal componets 
    14. eigs:<array, (nw)>: Eigenvalues
    15. full_eofs:<array, (nw, np)>: EOF with invalid points being filled
    
    
    Functions
    --------------------------------------------------------------------------
    1. __init__: Initialization
    2. get_pcs:  Get principal components
    3. get_eofs: Get EOF functions 
    4. get_eigs: Get eigenvalues for covariance matrix 
    5. project_field: project field to EOFs 
    6. show_pcs: plot PCs
    7. show_eofs: plot EOFs 
    8. show_eof_map: plot EOF over a map 
    

    



    
    
    
    """
    def __init__(self, data, wgt=None, filling_val=None,\
                     remove_mean=False, scaling_factor=None):
        """
        Inputs:
        -------------------------------------------
        1. data:<array, (nt, nx, (ny),...)>: original temporal spatial data
        2. wgt:<array, (nt, , nx, (ny),...)>: weigthing factor
        3. remove_mean:<T/F>: if ture, the mean of each column will be removed. 
        4. scaling_factor:<float>: factor using to scaling
        
        """
        # S1: get data shape

        self.dims=npy.shape(data)
        self.ndims=len(self.dims)
        
        # S2: make two copies of data and weight
        

        self.data = npy.array(data)
        
        if (wgt<>None):
            self.wgt=npy.array(wgt)
            self.usd_wgt=npy.array(wgt)
        else:
            self.wgt=None
            self.usd_wgt=None
        
        self.usd_data=npy.array(data)
        self.dims_x=self.dims[1:]
        
        # S2: reshape usd_data and usd_wgt to two dimensions if necessary
        # #c: usd_data=[nt, nx*ny*...]
        # #c: usd_wgt=[nt, nx*ny*...]
        
        if (filling_val==None):
            try:
                filling_val=getattr(data, 'missing_value')
            except:
                filling_val=-999.0
        
        self.filling_val=filling_val
        self.do_reshape=False
        
        if (self.ndims>2):
            
            self.do_reshape=True
            self.usd_data=npy.reshape(self.usd_data, [self.dims[0], -1])
            
            if (self.usd_wgt<>None):
                self.usd_wgt=npy.reshape(self.usd_wgt, [self.dims[0], -1])
            

                
        # #c: keep the shape 
        
        self.nt, self.np=npy.shape(self.usd_data)
        
        # S3: removing points with filling values
        
        usd_idx=npy.where(self.usd_data[0,:]<>filling_val)
        self.usd_idx=npy.squeeze(usd_idx)
        self.usd_data=self.usd_data[:, self.usd_idx]
        if (self.usd_wgt<>None):
            self.usd_wgt=self.usd_wgt[:, self.usd_idx]
    
            
        if (remove_mean):
            mean_val=npy.mean(self.usd_data, axis=0)
            self.usd_data=self.usd_data[:,:]-mean_val[npy.newaxis, :]
            
        # S4: SVD. 
        # self.w is the sqrt of eigenvalues of (usd_data^T)*usd_data
        
        if (self.usd_wgt<>None):
            usd_data=self.usd_data*self.usd_wgt
        else:
            usd_data=self.usd_data
        
        # #c: u is in shape of (nt, nw); vt is in shape of (nw, usd_np)
        
        
        self.u, self.w, self.vt=alg.svd(usd_data,  full_matrices=False)
        
        self.nw, self.usd_np=npy.shape(self.vt)
        
        # S5: construct eof including missing points 
        # #c: each row of eof is a eigenstate of covariance matrix 
        
        self.full_eofs=npy.zeros([self.nw, self.np], float)
        
        self.full_eofs[:,:]=filling_val
        self.full_eofs[:,self.usd_idx]=self.vt[:,:]
        
        # S6: construct eigen state
        
        if (scaling_factor==None):
            scaling_factor=1.0/(self.nt-1)
        
        # S3: eigenvalues of covariance matrix 
        
        self.eigs=scaling_factor*self.w*self.w
        
        # S7: construct Principal Components  (time function)
        
        
        self.pcs=self.u*self.w 
        
    def get_pcs(self, npcs=None):
        """
        
        Get principal components
        
        Inputs:
        -----------------------------
        1. npcs:<int>: number of principal componets  
        
        Returns:
        -----------------------------------------
        1. sel_pcs:<array, (nt, neofs)>: Principal components 

        """

        if (npcs==None):
            npcs=self.nw
            
        sel_pcs=self.pcs[:, 0:npcs]
        return sel_pcs

    def get_eofs(self, neofs=None):
        
        """
        Get EOF functions 
        
        Inputs:
        ==========================================
        1. neofs:<int>: number of selected eofs
       
        Returns:
        ============================================
        1. sel_eofs:<array, (neofs, nx,(ny), ...)>: EOFs
        
        
        
        """
        if (neofs==None):
            neofs=self.nw
        
        sel_eofs=self.full_eofs[0:neofs, :]
        if (self.do_reshape):
            eof_shape=[neofs]+self.dims[1:]
            sel_eofs=npy.reshape(sel_eofs, [eof_shape])
        
        

        return sel_eofs
    

    
    def get_eigs(self, neigs=None):
    
        """
        Get eigenvalues for covariance matrix 
        
        Inputs:
        ---------------------------------------------
        1. neigs:<int>: number of eigenvalues
        

        Returns:
        ----------------------------------------------
        1. sel_eigs:<array, (neigs)>: eigenvalue of covariance matrix 
        """
        if (neigs==None):
            neigs=self.nw

        sel_eigs=self.eigs[0:neigs]
        return sel_eigs
    
    def project_field(self, field, neofs=None, wgt=None, remove_mean=True, filling_val=None):
        """
        project field to EOFs 
        
        Inputs:
        -------------------------------------------------------
        1. field:<array, (nt, nx, (ny)...>: time-temporal data
        2. neofs:<int>: number of EOFs
        3. wgt:<array, (nt, nx, (ny)...>: weighting factor
        4. remove_mean:<T/F>: True, if mean_values to be removed from input data
        5. filling_val:<float>: filling values
        
        
        Returns:
        --------------------------------------------------------
        1. proj_pcs:<array, (nt, neofs)>: projection of fields over EOFs 
        
        """

        # S1: check dimension 
        
        if (filling_val==None):
            filling_val=self.filling_val
        
        if (neofs==None):
            neofs=self.nw
        
        fld_dims=npy.shape(field)
        
        if (len(fld_dims)<>self.ndims):
            print 'Shape of input field is different from shape of EOFs:' , fld_dims, dims
            
            return None

        
        for ix in range(1, self.ndims):
            if (fld_dims[ix]<>self.dims[ix]):
                print 'Shape of input field is different from shape of EOFs:' , fld_dims, dims
                return None
        
        usd_fld=npy.array(field)

        if (self.do_reshape):
            usd_fld=npy.reshape(usd_fld, [fld_dims[0], -1])

            
        usd_idx=npy.where(usd_fld[0,:]<>filling_val)
        usd_idx=npy.squeeze(usd_idx)
        usd_np=npy.size(usd_idx)
        if (usd_np<>self.usd_np):
            print 'Number of valid data points for input fields is  different from original data:', usd_np, self.usd_np
            return None
        
        dp=usd_np-self.usd_np

        idx=npy.where(dp<>0)
        idx=npy.squeeze(idx)
        if (npy.size(idx)>0):
            print 'Valid points for input field are different from the original data:', idx
            return None
        
        usd_fld=usd_fld[:, usd_idx]
        
        if (remove_mean):
            mean_val=npy.mean(usd_fld, axis=0)
            usd_fld[:,:]=usd_fld[:,:]-mean_val[npy.newaxis, :]
            
        if (wgt<>None):
            usd_wgt=npy.array(usd_wgt)
            if (self.do_reshape):
                usd_wgt=npy.reshape(usd_wgt, [fld_dims[0], -1])
            usd_wgt=usd_wgt[:, usd_idx]
        
            usd_fld=usd_fld*usd_wgt
        

        # projections 
        sel_eofs=self.vt[0:neofs, :]
        proj_pcs=npy.dot(usd_fld, npy.transpose(sel_eofs))
        
        return proj_pcs
    
        
    def show_pcs(self, ipc_lst, tau=None):
        """
        
        Display Principal components
        
        Inputs:
        ------------------------------
        1. ipc_lst:<list, t:int>: selected PCS
        2. tau:<float>: time array
        
        
        """
        
        ax=plt.subplot(111)
        if (type(ipc_lst)==type(1)):
            ipc=ipc_lst
            sel_pc=self.pcs[:,ipc]
            if (tau<>None):
                plt.plot(tau, sel_pc)
            else:
                plt.plot(sel_pc)
        else:
            for ipc in ipc_lst:
                
                sel_pc=self.pcs[:,ipc]
                if (tau<>None):
                    plt.plot(tau, sel_pc)
                else:
                    plt.plot(sel_pc)
        
        stitle='The first %3.3d PCs' % (npy.size(ipc_lst))
        
        plt.title(stitle)
        ax.set_xlabel('Time')
        ax.set_ylabel('')
        
        plt.show()
        

    def show_eofs(self, ieof_lst, x=None):
        
        """
        
        Display EOFs
        
        Inputs:
        ------------------------------
        1. ieof_lst:<list, t:int>: selected EOFs
        2. x:<float>: spatial array
                
        """
        
        ax=plt.subplot(111)
        if (type(ieof_lst)==type(1)):
            ieof=ieof_lst
            sel_eof=self.vt[ieof,:]
            if (x<>None):
                plt.plot(x, sel_eof)
            else:
                plt.plot(sel_eof)
        else:
            for ieof in ieof_lst:
                sel_eof=self.vt[ieof,:]
                if (x<>None):
                    plt.plot(x, sel_eof)
                else:
                    plt.plot(sel_eof)
                    
        
        stitle='The first %3.3d EOFs' % (npy.size(ieof_lst))
        
        plt.title(stitle)
        ax.set_xlabel('X')
        ax.set_ylabel('')
        
        plt.show()
        
    def show_eof_map(self, ieof, rlon, rlat, **keywords):
        """
        show selected EOF over a map 
        """
        sel_eof=self.full_eofs[ieof, ...]
                

        if (self.do_reshape):
            sel_eof=npy.reshape(sel_eof, self.dims[1:])
        
        
        
        nx, ny=npy.shape(sel_eof)
        if (nx<>npy.size(rlon)):
            sel_eof=npy.transpose(sel_eof)
        
        sel_eof=ma.masked_where(sel_eof==self.filling_val, sel_eof)
         
        gpl.plot_map(sel_eof, rlon=rlon, rlat=rlat, **keywords)
        plt.show()
        
    def get_variance_fraction(self, neigs):
        """
    
        Getn variance 
        Inputs:
        ----------------------------
        1. neigs:<int>: number of eigenvalues

        Returns:
        ------------------------------
        1. wgt:<float>: fraction of selected eigenvalues
        
        
        """
        sel_eig=npy.sum(self.eigs[0:neigs])
        wgt=sel_eig/npy.sum(self.eigs)
        return wgt
    
        
if (__name__ == "__main__"):

    print 'see test_eof.py'
    
        
    
    
